@@ -1,6 +1,7 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
 from decimal import Decimal
+from itertools import groupby
 
 from sql import Null
 from sql.aggregate import Sum
@@ -9,15 +10,15 @@ from sql.functions import Abs
 
 from trytond import backend
 from trytond.i18n import gettext
-from trytond.pool import Pool, PoolMeta
 from trytond.model import ModelSQL, ModelView, fields
-from trytond.pyson import Eval, Id
-from trytond.wizard import (Wizard, StateView, StateAction, StateTransition,
-    Button)
-from trytond.transaction import Transaction
-from trytond.tools.multivalue import migrate_property
 from trytond.modules.company.model import CompanyValueMixin
 from trytond.modules.currency.fields import Monetary
+from trytond.pool import Pool, PoolMeta
+from trytond.pyson import Bool, Eval, Id, If
+from trytond.tools.multivalue import migrate_property
+from trytond.transaction import Transaction
+from trytond.wizard import (
+    Button, StateAction, StateTransition, StateView, Wizard)
 
 from .exceptions import BlockedWarning, GroupWarning
 from .payment import KINDS
@@ -30,8 +31,7 @@ class MoveLine(metaclass=PoolMeta):
             currency='payment_currency', digits='payment_currency',
             states={
                 'invisible': ~Eval('payment_kind'),
-                },
-            depends=['payment_kind']),
+                }),
         'get_payment_amount', searcher='search_payment_amount')
     payment_currency = fields.Function(fields.Many2One(
             'currency.currency', "Payment Currency"),
@@ -40,8 +40,7 @@ class MoveLine(metaclass=PoolMeta):
         readonly=True,
         states={
             'invisible': ~Eval('payment_kind'),
-            },
-        depends=['payment_kind'])
+            })
     payment_kind = fields.Function(fields.Selection([
                 (None, ''),
                 ] + KINDS, 'Payment Kind'), 'get_payment_kind')
@@ -52,7 +51,6 @@ class MoveLine(metaclass=PoolMeta):
                 (Eval('payment_kind') == 'payable')
                 & ((Eval('credit', 0) > 0) | (Eval('debit', 0) < 0))),
             },
-        depends=['payment_kind', 'debit', 'credit'],
         help="Check if the line will be paid by direct debit.")
 
     @classmethod
@@ -243,6 +241,7 @@ class CreateDirectDebit(Wizard):
         pool = Pool()
         Line = pool.get('account.move.line')
         payments = Line.pay_direct_debit(date=self.start.date)
+        action['domains'] = []
         return action, {
             'res_id': [p.id for p in payments],
             }
@@ -279,8 +278,7 @@ class PayLineAskJournal(ModelView):
         required=True, domain=[
             ('company', '=', Eval('company', -1)),
             ('currency', '=', Eval('currency', -1)),
-            ],
-        depends=['company', 'currency'])
+            ])
 
     journals = fields.One2Many(
         'account.payment.journal', None, 'Journals', readonly=True)
@@ -422,10 +420,7 @@ class PayLine(Wizard):
             amount=line.payment_amount,
             line=line,
             )
-        date = self.start.date or line.maturity_date
-        # Use default value when empty
-        if date:
-            payment.date = date
+        payment.date = self.start.date or line.maturity_date
         return payment
 
     def do_pay(self, action):
@@ -478,8 +473,7 @@ class ConfigurationPaymentGroupSequence(ModelSQL, CompanyValueMixin):
             ('company', 'in', [Eval('company', -1), None]),
             ('sequence_type', '=',
                 Id('account_payment', 'sequence_type_account_payment_group')),
-            ],
-        depends=['company'])
+            ])
 
     @classmethod
     def __register__(cls, module_name):
@@ -518,7 +512,6 @@ class Invoice(metaclass=PoolMeta):
             'invisible': Eval('type') != 'in',
             'readonly': Eval('state') != 'draft',
             },
-        depends=['type', 'state'],
         help="Check if the invoice is paid by direct debit.")
 
     @classmethod
@@ -543,25 +536,103 @@ class Invoice(metaclass=PoolMeta):
         Date = pool.get('ir.date')
         context = Transaction().context
 
-        today = Date.today()
-
         amounts = super(Invoice, cls).get_amount_to_pay(invoices, name)
 
         if context.get('with_payment', True):
-            for invoice in invoices:
-                for line in invoice.lines_to_pay:
-                    if line.reconciliation:
-                        continue
-                    if (name == 'amount_to_pay_today'
-                            and line.maturity_date
-                            and line.maturity_date > today):
-                        continue
-                    payment_amount = Decimal(0)
-                    for payment in line.payments:
-                        if payment.state != 'failed':
+            for company, c_invoices in groupby(
+                    invoices, key=lambda i: i.company):
+                with Transaction().set_context(company=company.id):
+                    today = Date.today()
+                for invoice in c_invoices:
+                    for line in invoice.lines_to_pay:
+                        if line.reconciliation:
+                            continue
+                        if (name == 'amount_to_pay_today'
+                                and line.maturity_date
+                                and line.maturity_date > today):
+                            continue
+                        payment_amount = Decimal(0)
+                        for payment in line.payments:
                             with Transaction().set_context(date=payment.date):
                                 payment_amount += Currency.compute(
-                                    payment.currency, payment.amount,
+                                    payment.currency, payment.amount_line_paid,
                                     invoice.currency)
-                    amounts[invoice.id] -= payment_amount
+                        amounts[invoice.id] -= payment_amount
         return amounts
+
+
+class StatementLine(metaclass=PoolMeta):
+    __name__ = 'account.statement.line'
+
+    @classmethod
+    def __setup__(cls):
+        super().__setup__()
+        cls.related_to.domain['account.payment'] = [
+            ('company', '=', Eval('company', -1)),
+            If(Bool(Eval('party')),
+                ('party', '=', Eval('party')),
+                ()),
+            ('state', 'in', ['processing', 'succeeded', 'failed']),
+            ('currency', '=', Eval('currency', -1)),
+            ('kind', '=',
+                If(Eval('amount', 0) > 0, 'receivable',
+                    If(Eval('amount', 0) < 0, 'payable', ''))),
+            ]
+        cls.related_to.search_order['account.payment'] = [
+            ('amount', 'ASC'),
+            ('state', 'ASC'),
+            ]
+        cls.related_to.search_context.update({
+                'amount_order': Eval('amount', 0),
+                })
+
+    @classmethod
+    def _get_relations(cls):
+        return super()._get_relations() + ['account.payment']
+
+    @property
+    @fields.depends('related_to')
+    def payment(self):
+        pool = Pool()
+        Payment = pool.get('account.payment')
+        related_to = getattr(self, 'related_to', None)
+        if isinstance(related_to, Payment) and related_to.id >= 0:
+            return related_to
+
+    @payment.setter
+    def payment(self, value):
+        self.related_to = value
+
+    @fields.depends(
+        'party', 'statement', '_parent_statement.journal',
+        methods=['payment'])
+    def on_change_related_to(self):
+        super().on_change_related_to()
+        if self.payment:
+            if not self.party:
+                self.party = self.payment.party
+            if self.payment.line:
+                self.account = self.payment.line.account
+
+
+class Dunning(metaclass=PoolMeta):
+    __name__ = 'account.dunning'
+
+    def get_active(self, name):
+        return super().get_active(name) and self.line.payment_amount > 0
+
+    @classmethod
+    def search_active(cls, name, clause):
+        if tuple(clause[1:]) in {('=', True), ('!=', False)}:
+            domain = ('line.payment_amount', '>', 0)
+        elif tuple(clause[1:]) in {('=', False), ('!=', True)}:
+            domain = ('line.payment_amount', '<=', 0)
+        else:
+            domain = []
+        return [super().search_active(name, clause), domain]
+
+    @classmethod
+    def _overdue_line_domain(cls, date):
+        return [super()._overdue_line_domain(date),
+            ('payment_amount', '>', 0),
+            ]
