@@ -4,22 +4,24 @@ from collections import defaultdict
 from decimal import Decimal
 from itertools import groupby
 
-from sql.aggregate import Sum, Count
+from sql.aggregate import Count, Sum
+from sql.functions import CharLength
+from sql.operators import Abs
 
 from trytond.i18n import gettext
 from trytond.model import (
-    Workflow, ModelView, ModelSQL, DeactivableMixin, fields)
+    DeactivableMixin, Index, ModelSQL, ModelView, Workflow, fields)
 from trytond.model.exceptions import AccessError
+from trytond.modules.company.model import (
+    employee_field, reset_employee, set_employee)
+from trytond.modules.currency.fields import Monetary
+from trytond.pool import Pool
 from trytond.pyson import Eval, If
 from trytond.rpc import RPC
-from trytond.tools import reduce_ids, grouped_slice, cursor_dict
+from trytond.tools import (
+    cursor_dict, grouped_slice, reduce_ids, sortable_values)
 from trytond.transaction import Transaction
-from trytond.tools import sortable_values
-from trytond.wizard import Wizard, StateView, StateAction, Button
-from trytond.pool import Pool
-from trytond.modules.company.model import (
-    employee_field, set_employee, reset_employee)
-from trytond.modules.currency.fields import Monetary
+from trytond.wizard import Button, StateAction, StateView, Wizard
 
 from .exceptions import OverpayWarning
 
@@ -34,8 +36,7 @@ class Journal(DeactivableMixin, ModelSQL, ModelView):
     __name__ = 'account.payment.journal'
     name = fields.Char('Name', required=True)
     currency = fields.Many2One('currency.currency', 'Currency', required=True)
-    company = fields.Many2One('company.company', 'Company', required=True,
-        select=True)
+    company = fields.Many2One('company.company', "Company", required=True)
     process_method = fields.Selection([
             ('manual', 'Manual'),
             ], 'Process Method', required=True)
@@ -64,23 +65,21 @@ class Group(ModelSQL, ModelView):
     number = fields.Char('Number', required=True, readonly=True)
     company = fields.Many2One(
         'company.company', "Company",
-        required=True, readonly=True, select=True)
+        required=True, readonly=True)
     journal = fields.Many2One('account.payment.journal', 'Journal',
         required=True, readonly=True, domain=[
             ('company', '=', Eval('company', -1)),
-            ],
-        depends=['company'])
+            ])
     kind = fields.Selection(KINDS, 'Kind', required=True, readonly=True)
     payments = fields.One2Many(
         'account.payment', 'group', 'Payments', readonly=True,
         domain=[
             ('company', '=', Eval('company', -1)),
             ('journal', '=', Eval('journal', -1)),
-            ],
-        depends=['company', 'journal'])
+            ])
     currency = fields.Function(fields.Many2One(
             'currency.currency', "Currency"),
-        'on_change_with_currency')
+        'on_change_with_currency', searcher='search_currency')
     payment_count = fields.Function(fields.Integer(
             "Payment Count",
             help="The number of payments in the group."),
@@ -121,6 +120,11 @@ class Group(ModelSQL, ModelView):
             table_h.column_rename('reference', 'number')
         super(Group, cls).__register__(module_name)
 
+    @classmethod
+    def order_number(cls, tables):
+        table, _ = tables[None]
+        return [CharLength(table.number), table.number]
+
     @staticmethod
     def default_company():
         return Transaction().context.get('company')
@@ -268,39 +272,45 @@ class Group(ModelSQL, ModelView):
         if self.journal:
             return self.journal.currency.id
 
+    @classmethod
+    def search_currency(cls, name, clause):
+        return [('journal.' + clause[0],) + tuple(clause[1:])]
+
 
 _STATES = {
     'readonly': Eval('state') != 'draft',
     }
-_DEPENDS = ['state']
 
 
 class Payment(Workflow, ModelSQL, ModelView):
     'Payment'
     __name__ = 'account.payment'
     company = fields.Many2One(
-        'company.company', "Company", required=True, select=True,
-        states=_STATES, depends=_DEPENDS)
+        'company.company', "Company", required=True, states=_STATES)
     journal = fields.Many2One('account.payment.journal', 'Journal',
         required=True, states=_STATES, domain=[
             ('company', '=', Eval('company', -1)),
-            ],
-        depends=_DEPENDS + ['company'])
+            ])
     currency = fields.Function(fields.Many2One('currency.currency',
             'Currency'), 'on_change_with_currency',
         searcher='search_currency')
     kind = fields.Selection(KINDS, 'Kind', required=True,
-        states=_STATES, depends=_DEPENDS)
+        states=_STATES)
     party = fields.Many2One(
         'party.party', "Party", required=True, states=_STATES,
         context={
             'company': Eval('company', -1),
             },
-        depends=_DEPENDS + ['company'])
-    date = fields.Date('Date', required=True, states=_STATES, depends=_DEPENDS)
+        depends={'company'})
+    date = fields.Date('Date', required=True, states=_STATES)
     amount = Monetary(
         "Amount", currency='currency', digits='currency', required=True,
-        states=_STATES, depends=_DEPENDS)
+        states={
+            'readonly': ~Eval('state').in_(
+                If(Eval('process_method') == 'manual',
+                    ['draft', 'processing'],
+                    ['draft'])),
+            })
     line = fields.Many2One('account.move.line', 'Line', ondelete='RESTRICT',
         domain=[
             ('move.company', '=', Eval('company', -1)),
@@ -316,6 +326,7 @@ class Payment(Workflow, ModelSQL, ModelView):
             If(Eval('state') == 'draft',
                 [
                     ('reconciliation', '=', None),
+                    ('maturity_date', '!=', None),
                     ],
                 []),
             ['OR',
@@ -327,15 +338,13 @@ class Payment(Workflow, ModelSQL, ModelView):
                 ],
             ('move_state', '=', 'posted'),
             ],
-        states=_STATES, depends=_DEPENDS + ['party', 'currency', 'kind',
-            'company'])
-    description = fields.Char('Description', states=_STATES, depends=_DEPENDS)
+        states=_STATES)
+    description = fields.Char('Description', states=_STATES)
     origin = fields.Reference(
-        "Origin", selection='get_origin', select=True,
+        "Origin", selection='get_origin',
         states={
             'readonly': Eval('state') != 'draft',
-            },
-        depends=['state'])
+            })
     group = fields.Many2One('account.payment.group', 'Group', readonly=True,
         ondelete='RESTRICT',
         states={
@@ -343,11 +352,13 @@ class Payment(Workflow, ModelSQL, ModelView):
             },
         domain=[
             ('company', '=', Eval('company', -1)),
-            ],
-        depends=['state', 'company'])
+            ])
     process_method = fields.Function(
         fields.Selection('get_process_methods', "Process Method"),
         'on_change_with_process_method', searcher='search_process_method')
+    submitted_by = employee_field(
+        "Submitted by",
+        states=['submitted', 'processing', 'succeeded', 'failed'])
     approved_by = employee_field(
         "Approved by",
         states=['approved', 'processing', 'succeeded', 'failed'])
@@ -358,21 +369,46 @@ class Payment(Workflow, ModelSQL, ModelView):
         states=['failed', 'processing'])
     state = fields.Selection([
             ('draft', 'Draft'),
+            ('submitted', "Submitted"),
             ('approved', 'Approved'),
             ('processing', 'Processing'),
             ('succeeded', 'Succeeded'),
             ('failed', 'Failed'),
-            ], 'State', readonly=True, select=True)
+            ], "State", readonly=True, sort=False,
+        domain=[
+            If(Eval('kind') == 'receivable',
+                ('state', '!=', 'approved'),
+                ()),
+            ])
+
+    @property
+    def amount_line_paid(self):
+        if self.state != 'failed':
+            if self.line.second_currency:
+                payment_amount = abs(self.line.amount_second_currency)
+            else:
+                payment_amount = abs(self.line.credit - self.line.debit)
+            return max(min(self.amount, payment_amount), 0)
+        return Decimal(0)
 
     @classmethod
     def __setup__(cls):
         super(Payment, cls).__setup__()
+        t = cls.__table__()
+        cls._sql_indexes.add(
+            Index(
+                t, (t.state, Index.Equality()),
+                where=t.state.in_([
+                        'draft', 'submitted', 'approved', 'processing'])))
         cls._order.insert(0, ('date', 'DESC'))
         cls._transitions |= set((
-                ('draft', 'approved'),
+                ('draft', 'submitted'),
+                ('submitted', 'approved'),
+                ('submitted', 'processing'),
                 ('approved', 'processing'),
                 ('processing', 'succeeded'),
                 ('processing', 'failed'),
+                ('submitted', 'draft'),
                 ('approved', 'draft'),
                 ('succeeded', 'failed'),
                 ('succeeded', 'processing'),
@@ -381,15 +417,29 @@ class Payment(Workflow, ModelSQL, ModelView):
                 ))
         cls._buttons.update({
                 'draft': {
-                    'invisible': Eval('state') != 'approved',
+                    'invisible': ~Eval('state').in_(['submitted', 'approved']),
                     'icon': 'tryton-back',
                     'depends': ['state'],
                     },
-                'approve': {
+                'submit': {
                     'invisible': Eval('state') != 'draft',
                     'icon': 'tryton-forward',
                     'depends': ['state'],
                     },
+                'approve': {
+                    'invisible': (
+                        (Eval('state') != 'submitted')
+                        | (Eval('kind') == 'receivable')),
+                    'icon': 'tryton-forward',
+                    'depends': ['state', 'kind'],
+                    },
+                'proceed': {
+                    'invisible': (
+                        ~Eval('state').in_(['succeeded', 'failed'])
+                        | (Eval('process_method') != 'manual')),
+                    'icon': 'tryton-back',
+                    'depends': ['state', 'process_method'],
+                    },
                 'succeed': {
                     'invisible': ~Eval('state').in_(
                         ['processing', 'failed']),
@@ -435,6 +485,16 @@ class Payment(Workflow, ModelSQL, ModelView):
     def search_currency(cls, name, clause):
         return [('journal.' + clause[0],) + tuple(clause[1:])]
 
+    @classmethod
+    def order_amount(cls, tables):
+        table, _ = tables[None]
+        context = Transaction().context
+        column = cls.amount.sql_column(table)
+        if isinstance(context.get('amount_order'), Decimal):
+            return [Abs(column - abs(context['amount_order']))]
+        else:
+            return [column]
+
     @fields.depends('kind')
     def on_change_kind(self):
         self.line = None
@@ -502,10 +562,17 @@ class Payment(Workflow, ModelSQL, ModelView):
     @classmethod
     @ModelView.button
     @Workflow.transition('draft')
-    @reset_employee('approved_by', 'succeeded_by', 'failed_by')
+    @reset_employee('submitted_by', 'approved_by', 'succeeded_by', 'failed_by')
     def draft(cls, payments):
         pass
 
+    @classmethod
+    @ModelView.button
+    @Workflow.transition('submitted')
+    @set_employee('submitted_by')
+    def submit(cls, payments):
+        pass
+
     @classmethod
     @ModelView.button
     @Workflow.transition('approved')
@@ -534,6 +601,7 @@ class Payment(Workflow, ModelSQL, ModelView):
             return group
 
     @classmethod
+    @ModelView.button
     @Workflow.transition('processing')
     def proceed(cls, payments):
         assert all(p.group for p in payments)
@@ -586,6 +654,11 @@ class ProcessPayment(Wizard):
         Warning = pool.get('res.user.warning')
         payments = self.records
 
+        payments = [
+            p for p in payments
+            if p.state == 'approved'
+            or (p.state == 'submitted' and p.kind == 'receivable')]
+
         for payment in payments:
             if payment.line and payment.line.payment_amount < 0:
                 if Warning.check(str(payment)):
